import wandb
import time
import pandas as pd

from .utils import get_lr
from .logger import Logger


class Wandb:
    def __init__(self, config):
        assert (
            config.wandb_project != None
        ), "You need to specify a project name in the config to be able to use WandB (config.wandb_project='my_project_name')"
        self.config = config
        self.logger = Logger("WandB", config)
        self.start = -1

    def init(self, model, optimizer, criterion, scheduler):
        self.scheduler = scheduler

        run_name = self.get_run_name(optimizer, scheduler)

        wandb_config = self.config.__dict__
        classes_config = {
            "model": model.__class__.__name__,
            "optimizer": optimizer.__class__.__name__,
            "criterion": criterion.__class__.__name__,
        }
        if scheduler != None:
            classes_config["scheduler"] = scheduler.__class__.__name__
        wandb_config.update(classes_config)

        self.logger.log(
            f"Initializing wandb with project_name {self.config.wandb_project} and run name {run_name}"
        )

        wandb.init(
            project=self.config.wandb_project,
            name=run_name,
            config=wandb_config,
        )
        wandb.define_metric("iter")
        wandb.define_metric("time")
        wandb.define_metric("eval", step_metric="iter")

    def log(self, stats):
        if self.start == -1:
            self.start = time.time()

        payload = dict(**stats)
        payload["lr"] = get_lr(self.config, self.scheduler)
        payload["time"] = time.time() - self.start
        payload = pd.json_normalize(payload, sep="/")
        payload = payload.to_dict(orient="records")[0]

        wandb.log(payload)

    def get_run_name(self, optimizer, scheduler):
        optim_name = optimizer.__class__.__name__
        sched_name = scheduler.__class__.__name__ if scheduler != None else "None"
        name = f"{self.config.name}_lr={self.config.lr}_bs={self.config.batch_size}_optim={optim_name}_sched={sched_name}"
        return name



import torch
import torch.nn as nn
from torch.optim import AdamW

from .utils import Record


def Config(
    name,
    max_iterations,
    max_eval_iterations=None,
    device="auto",
    lr=1e-4,
    epochs=None,
    batch_size=64,
    compile=True,
    stats_every=1,
    evaluate=True,
    evaluate_every=10,
    eval_batch_size=None,
    metrics=[],
    grad_norm_clip=1.0,
    num_workers=4,
    eval_num_workers=0,
    optimizer=AdamW,
    criterion=nn.MSELoss,
    scheduler=None,
    optimizer_params=None,
    scheduler_params=None,
    wandb=True,
    wandb_project=None,
    checkpoint_folder="./checkpoints",
    verbose=False,
    **kwargs
):
    """
    Returns a configuration object with the specified hyperparameters.

    Args:
        name (str): Name of the configuration.
        max_iterations (int): Maximum number of iterations.
        max_eval_iterations (int, optional): Maximum number of iterations for the evaluation dataset. Defaults to None.
        device (str, optional): Device to train on. Defaults to "auto" which will use GPU if available.
        lr (float, optional): Learning rate. Defaults to 1e-4.
        epochs (int, optional): Number of epochs (if max_iterations is not defined). Defaults to None.
        batch_size (int, optional): Batch size. Defaults to 64.
        compile (bool, optional): Whether to compile the model, this can significantly improve training time but is not supported on all GPUs. Defaults to True.
        stats_every (int, optional): Compute statistics every `stats_every` iterations. Defaults to 1.
        evaluate (bool or str, optional): Whether to periodically evaluate the model on the evaluation dataset, or 'epoch' to evaluate every epoch. Defaults to True.
        evaluate_every (int, optional): Evaluate every `evaluate_every` iterations / or epoch if evaluate is set to 'epoch'. Defaults to 10.
        eval_batch_size (int, optional): Batch size for the evaluation dataset. Defaults to None in which case it will be equal to the training batch size.
        metrics (list, optional): List of metric classes to compute. Defaults to []. Possible values: metrics.Accuracy, metrics.RMSE, or your own class inheriting Metrics.
        grad_norm_clip (float, optional): Gradient norm clipping. Defaults to 1.0.
        num_workers (int, optional): Number of workers for the dataloader. Defaults to 4.
        eval_num_workers (int, optional): Number of workers for the evaluation dataloader. Note that a value > 0 can cause an AssertionError: 'can only test a child process during evaluation'. Defaults to 0.
        optimizer (torch.optim.Optimizer, optional): Optimizer. Defaults to AdamW.
        criterion (torch.nn.modules.loss._Loss, optional): Loss function. Defaults to nn.MSELoss.
        scheduler (torch.optim.lr_scheduler._LRScheduler, optional): Scheduler. Defaults to None.
        optimizer_params (dict, optional): Optimizer parameters. Defaults to None.
        scheduler_params (dict, optional): Scheduler parameters. Defaults to None.
        wandb (bool, optional): Whether to use wandb. Defaults to True.
        wandb_project (str, optional): Wandb project name, if wandb is set to True. Defaults to None.
        checkpoint_folder (str, optional): Folder to save checkpoints. Defaults to "./checkpoints".
        verbose (bool, optional): Whether to print details of whats going on in the system. Defaults to False.

    Returns:
        Record: Configuration object with the specified hyperparameters.
    """
    if optimizer_params == None:
        optimizer_params = {}

    if scheduler == None:
        scheduler = torch.optim.lr_scheduler.OneCycleLR
        scheduler_params = dict(
            {
                "max_lr": lr,
                "pct_start": 0.02,
                "anneal_strategy": "cos",
                "cycle_momentum": False,
                "div_factor": 1e2,
                "final_div_factor": 0.05,
                "total_steps": max_iterations,
            }
            if scheduler_params == None
            else scheduler_params
        )

    if eval_batch_size == None:
        eval_batch_size = batch_size

    return Record(
        name=name,
        max_iterations=max_iterations,
        max_eval_iterations=max_eval_iterations,
        device=device,
        lr=lr,
        epochs=epochs,
        batch_size=batch_size,
        compile=compile,
        stats_every=stats_every,
        evaluate=evaluate,
        evaluate_every=evaluate_every,
        eval_batch_size=eval_batch_size,
        metrics=metrics,
        grad_norm_clip=grad_norm_clip,
        num_workers=num_workers,
        eval_num_workers=eval_num_workers,
        optimizer=optimizer,
        criterion=criterion,
        scheduler=scheduler,
        optimizer_params=optimizer_params,
        scheduler_params=scheduler_params,
        wandb=wandb,
        wandb_project=wandb_project,
        checkpoint_folder=checkpoint_folder,
        verbose=verbose,
        **kwargs
    )



import torch
from torch.utils.data import DataLoader

from .metrics import Loss
from .utils import to_device, get_lr
from .logger import Logger


class Statistics:
    def __init__(self, model, criterion, scheduler, config, eval_dataset=None):
        assert (config.evaluate and eval_dataset != None) or (
            not config.evaluate
        ), "You have chosen to evaluate the model in the Config, but no evaluation dataset is passed"

        self.model = model
        self.criterion = criterion
        self.scheduler = scheduler
        self.config = config
        self.eval_dataset = eval_dataset

        self.lr = config.lr

        self.train_metrics = self.create_metrics()
        self.eval_metrics = self.create_metrics()

        self.logger = Logger("Statistics", config)

    def create_metrics(self):
        metrics = [m for m in self.config.metrics]
        metrics.append(Loss(self.criterion))
        return metrics

    @torch.no_grad()
    def evaluate(self):
        self.logger.log("Evaluating")
        self.model.eval()

        device = self.config.device
        metric_values = [[] for _ in range(len(self.eval_metrics))]

        self.eval_loader = DataLoader(
            self.eval_dataset,
            shuffle=False,
            pin_memory=self.config.device != "cpu",
            batch_size=self.config.eval_batch_size,
            num_workers=self.config.eval_num_workers,
        )

        for i, (data, target) in enumerate(self.eval_loader):
            if (
                self.config.max_eval_iterations != None
                and i >= self.config.max_eval_iterations
            ):
                break
            data, target = to_device(data, device), to_device(target, device)
            output = self.model(data)

            for i, metric in enumerate(self.eval_metrics):
                metric_value = metric.compute(output, target, evaluate=True)
                metric_values[i].append(metric_value)

        eval_stats = {}
        for metric, value in zip(self.eval_metrics, metric_values):
            stat = metric.update(value, evaluate=True)
            eval_stats.update(stat)

        self.model.train()

        return eval_stats

    def log_stats(self, stats, epoch, iteration):
        log = f"[{epoch:03d} | {iteration:05d}:{self.config.max_iterations:05d}]"
        for metric in self.train_metrics:
            log += f" | [{metric.name}] tr: {stats['train'][metric.name]:.4f}"
            if "eval" in stats:
                log += f", ev: {stats['eval'][metric.name]:.4f}"
        log += f" | [Lr] {stats['lr']:.4f}"
        self.logger.log(log)

    @torch.no_grad()
    def register(self, train_output, train_target, train_loss, epoch, iteration):
        # FIXME: move to device?

        train_stats = {}
        for metric in self.train_metrics:
            value = (
                train_loss
                if metric.name == "loss"
                else metric.compute(train_output, train_target, evaluate=False)
            )
            stat = metric.update(value, evaluate=False)
            train_stats.update(stat)

        self.lr = get_lr(self.config, self.scheduler)

        stats = {"epoch": epoch, "iter": iteration, "train": train_stats, "lr": self.lr}

        if self.config.evaluate != False:
            eval_epoch = (
                self.config.evaluate == "epoch"
                and epoch % self.config.evaluate_every == 0
            )
            eval_iter = (
                self.config.evaluate == True
                and iteration % self.config.evaluate_every == 0
            )
            if eval_epoch or eval_iter:
                eval_stats = self.evaluate()
                stats["eval"] = eval_stats

        if self.config.verbose:
            self.log_stats(stats, epoch, iteration)

        return stats



from .utils import get_date


class Logger:
    def __init__(self, name, config):
        self.name = name
        self.config = config

    def log(self, *args, force=False):
        if self.config.verbose or force:
            d = get_date()
            print(f"{d} | {self.name:<10} |", *args)



import os
import time
import torch
import torch.nn as nn
from torch.utils.data import DataLoader, RandomSampler

from .wandb import Wandb
from .stats import Statistics
from .utils import to_device, get_date, Callbacks
from .logger import Logger


class Trainer(Callbacks):
    def __init__(self, model, config):
        self.config = config
        self.logger = Logger("Trainer", config)
        self.date = get_date()

        if self.config.device == "auto":
            self.config.device = "cuda" if torch.cuda.is_available() else "cpu"

        self.logger.log(f"Set on device {self.config.device}", force=True)

        self.model = model.to(device=self.config.device)
        if self.config.compile:
            self.logger.log(f"Compiling model...", force=True)
            self.model = torch.compile(self.model)
            self.logger.log(f"Model compiled!", force=True)

        self.epoch = 0
        self.iteration = 0
        self.optimizer = self.config.optimizer(
            self.model.parameters(), lr=self.config.lr, **self.config.optimizer_params
        )
        self.criterion = self.config.criterion()
        self.scheduler = None
        if self.config.scheduler:
            self.scheduler = self.config.scheduler(
                self.optimizer, **self.config.scheduler_params
            )

        if config.wandb:
            self.wandb = Wandb(config)

        Callbacks.__init__(self)

    @staticmethod
    def get_file_path(date, epoch, iteration):
        return f"{date}_epoch={epoch:03d}_iter={iteration:06d}.pt"

    @staticmethod
    def get_checkpoint_path(config, date, epoch, iteration):
        folder = os.path.join(config.checkpoint_folder, config.name)
        file = Trainer.get_file_path(date, epoch, iteration)
        return folder, os.path.join(folder, file)

    def save_model(self):
        state = {
            "epoch": self.epoch,
            "iter": self.iteration,
            "config": self.config,
            "model": self.model.to(device="cpu").state_dict(),
            "optimizer": self.optimizer.state_dict(),
            "scheduler": self.scheduler.state_dict()
            if self.scheduler != None
            else None,
        }

        folder, cp_path = Trainer.get_checkpoint_path(
            self.config, self.date, self.epoch, self.iteration
        )
        if not os.path.exists(folder):
            os.makedirs(folder)

        self.logger.log(
            f"Saving model {self.config.name} at epoch {self.epoch}, iter {self.iteration} to {cp_path}",
            force=True,
        )
        torch.save(state, cp_path)
        return cp_path

    def _load_state_dict(self, checkpoint, keep_states):
        # self.epoch = checkpoint["epoch"]
        # self.iteration = checkpoint["iter"]
        self.model.load_state_dict(checkpoint["model"])
        if keep_states:
            self.optimizer.load_state_dict(checkpoint["optimizer"])
            if self.config.scheduler != None:
                self.scheduler.load_state_dict(checkpoint["scheduler"])

    @staticmethod
    def from_pretrained(model, config, cp_path, keep_states=True):
        Logger("Trainer", config).log("Loading checkpoint from", cp_path, force=True)
        checkpoint = torch.load(cp_path)
        # Don't use the checkpoint config as it contains erroneous data such as optimizer, scheduler represented as strings
        # Moreover, this allows to change the config between runs, even tho this is already possible from one training to another
        trainer = Trainer(model, config)
        trainer._load_state_dict(checkpoint, keep_states)
        return trainer

    def on_batch_end(self, statistics, output, target, loss):
        self.trigger_callbacks("on_batch_end")

        if self.config.stats_every == None or (
            self.iteration != 0 and self.iteration % self.config.stats_every != 0
        ):
            return

        stats = statistics.register(output, target, loss, self.epoch, self.iteration)

        if self.config.wandb:
            self.wandb.log(stats)

        self.trigger_callbacks("on_stats", **stats)

    def fit(self, train_dataset, eval_dataset=None):
        device = self.config.device
        self.model.train()
        self.date = get_date()

        statistics = Statistics(
            self.model,
            self.criterion,
            self.scheduler,
            self.config,
            eval_dataset,
        )

        train_loader = DataLoader(
            train_dataset,
            sampler=RandomSampler(train_dataset, replacement=True),
            shuffle=False,
            pin_memory=device != "cpu",
            batch_size=self.config.batch_size,
            num_workers=self.config.num_workers,
        )
        data_iter = iter(train_loader)

        if self.config.wandb:
            self.logger.log("Initializing wandb")
            self.wandb.init(self.model, self.optimizer, self.criterion, self.scheduler)

        self.logger.log("Starting training")
        while self.iteration < self.config.max_iterations and (
            self.config.epochs == None or self.epoch < self.config.epochs
        ):
            try:
                batch = next(data_iter)
            except StopIteration:
                data_iter = iter(train_loader)
                self.epoch += 1
                self.trigger_callbacks("on_epoch_end")
                batch = next(data_iter)

            self.trigger_callbacks("on_batch_start")
            data, target = batch
            data, target = to_device(data, device), to_device(target, device)

            output = self.model(data)
            loss = self.criterion(output, target)

            loss.backward()
            if self.config.grad_norm_clip != None and self.config.grad_norm_clip != 0.0:
                nn.utils.clip_grad_norm_(
                    self.model.parameters(), self.config.grad_norm_clip
                )
            self.optimizer.step()
            if self.scheduler:
                self.scheduler.step()
            self.optimizer.zero_grad(set_to_none=True)

            self.on_batch_end(statistics, output, target, loss)

            self.iteration += 1

        cp_path = self.save_model()
        return statistics, cp_path



import torch

# TODO: from sklearn.metrics import accuracy_score, precision_score, recall_score, f1_score
# FIXME: move to device?


class Metric:
    def __init__(self, name, mode):
        self.name = name
        self.mode = mode
        self.value = 0
        self.mode_value = float("inf") if mode == "min" else float("-inf")

    @torch.no_grad()
    def compute(self, output, target, **kwargs):
        """
        Computes the metric for the given output and target tensors.

        Args:
            output (torch.Tensor): The output tensor.
            target (torch.Tensor): The target tensor.
            kwargs (dict):
                evaluate: Whether in eval mode or not
                ...

        Returns:
            torch.Tensor: A tensor containing a single value corresponding to the computed metric.
        """
        raise NotImplementedError

    @torch.no_grad()
    def aggregate(self, value, **kwargs):
        if isinstance(value, tuple) or isinstance(value, list):
            return torch.stack(value).mean()
        else:
            return value

    @torch.no_grad()
    def update(self, value, **kwargs):
        value = self.aggregate(value, **kwargs)
        self.value = value.item()
        compare_func = min if self.mode == "min" else max
        self.mode_value = compare_func(self.mode_value, self.value)
        return {
            self.name: self.value,
            f"{self.mode}_{self.name}": self.mode_value,
        }


class Accuracy(Metric):
    def __init__(self):
        super().__init__("acc", "max")

    @torch.no_grad()
    def compute(self, output, target, **kwargs):
        pred = output.argmax(dim=1, keepdim=True)
        correct = torch.sum(pred == target)
        return correct * 100.0 / output.shape[0]


class RMSE(Metric):
    def __init__(self):
        super().__init__("rmse", "min")
        self.size = 0

    @torch.no_grad()
    def compute(self, output, target, **kwargs):
        self.size += output.shape[0]
        return torch.sum((target - output) ** 2)

    @torch.no_grad()
    def aggregate(self, value, **kwargs):
        if isinstance(value, tuple) or isinstance(value, list):
            value = torch.stack(value)
        mse_mean = torch.sum(value) / self.size
        self.size = 0
        return torch.sqrt(mse_mean)


class Loss(Metric):
    def __init__(self, criterion):
        super().__init__("loss", "min")
        self.criterion = criterion

    @torch.no_grad()
    def compute(self, output, target, **kwargs):
        return self.criterion(output, target)



from setuptools import setup, find_packages


with open("README.md", "r") as f:
    long_description = f.read()

setup(
    name="pyroml",
    version="0.0.11",
    author="Nathan Maire",
    author_email="nathan.maire@epfl.ch",
    description="Machine Learning tool allowing plug-and-play training for pytorch models",
    long_description=long_description,
    long_description_content_type="text/markdown",
    license="MIT",
    packages=find_packages(where=".", exclude=["tests"]),
    url="https://github.com/peacefulotter/pyroml",
    install_requires=[
        "torch",
        "pandas",
        "numpy",
        "wandb",
    ],
)



import time
import torch
from collections import defaultdict


def to_device(obj, device):
    if isinstance(obj, float) or isinstance(obj, int):
        return torch.tensor(obj).to(device)
    if torch.is_tensor(obj):
        return obj.to(device)
    if isinstance(obj, dict):
        return {k: to_device(v, device) for k, v in obj.items()}
    if isinstance(obj, tuple):
        return tuple(to_device(v, device) for v in obj)
    if isinstance(obj, list):
        return [to_device(v, device) for v in obj]
    return obj


def get_lr(config, scheduler):
    if scheduler == None:
        return config.lr
    return float(scheduler.get_last_lr()[0])


def get_date():
    return time.strftime("%Y-%m-%d_%H:%M", time.gmtime(time.time()))


class Record:
    def __init__(self, **kwargs):
        for k, v in kwargs.items():
            setattr(self, k, v)

    def __str__(self):
        return f"Config({str(self.__dict__)[1:-1]})"

    def __repr__(self):
        return self.__str__()


class Callbacks:
    def __init__(self):
        self.callbacks = defaultdict(list)

    def add_callback(self, onevent: str, callback):
        self.callbacks[onevent].append(callback)

    def set_callback(self, onevent: str, callback):
        self.callbacks[onevent] = [callback]

    def trigger_callbacks(self, onevent: str, **kwargs):
        for callback in self.callbacks.get(onevent, []):
            callback(self, **kwargs)
